/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.ships.arithmetic;

import java.util.Map;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.fleet.ships.tokenFifo.TokenFifoShipDefaults;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * @author ib27688
 *
 */
public abstract class ArithmeticTokenShip extends Component {
	InputTerminal in0;
	OutputTerminal ackIn0;	
	
	InputTerminal in1;
	OutputTerminal ackIn1;

	OutputTerminal tokenOut;
	InputTerminal ackTokenOut;

	
	OutputTerminal out;
	InputTerminal ackOut;
	InputTerminal tokenIn;
	OutputTerminal ackTokenIn;
	
	/* this mape will be used in getTerminal */
	Map<String, Terminal> terminalNameMap;
	
	ArithmeticAction computeAction;
	
	public ArithmeticTokenShip(String name, Parameters params) {
		super(name, null);
		
		Delay computeDelayParam= null;
		Delay ackDelayParam= null;
		Delay tokenDelayParam= null;
		Delay ackTokenDelayParam= null;
		
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(ArithmeticShipDefaults.COMPUTE_DELAY) ) {
					try {
						int d = Integer.parseInt(sValue);
						computeDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ ArithmeticShipDefaults.COMPUTE_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(ArithmeticShipDefaults.ACK_DELAY) ) {
					try {
						int d = Integer.parseInt(sValue);
						ackDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ ArithmeticShipDefaults.ACK_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(ArithmeticShipDefaults.TOKEN_DELAY) ) {
					try {
						int d = Integer.parseInt(sValue);
						tokenDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ ArithmeticShipDefaults.TOKEN_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(ArithmeticShipDefaults.ACK_TOKEN_DELAY) ) {
					try {
						int d = Integer.parseInt(sValue);
						ackTokenDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ ArithmeticShipDefaults.ACK_TOKEN_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}				
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				} // else
			} // for
		} // if params != null			
		
		buildArithmeticTokenShip();
		
		if (computeDelayParam != null) {
			computeAction.computeDataFn.setDelay(computeDelayParam);
		}
		if (ackDelayParam != null) {
			computeAction.ackFn0.setDelay(ackDelayParam);
			computeAction.ackFn1.setDelay(ackDelayParam);
		}
		if (tokenDelayParam != null) {
			computeAction.tokenOutFn.setDelay(tokenDelayParam);
		}
		if (ackTokenDelayParam != null) {
			computeAction.ackTokenInFn.setDelay(ackTokenDelayParam);
		}
		
	} // ArithmeticShip constructor

	protected void buildArithmeticTokenShip() {
		computeAction= new ArithmeticAction("ArithmeticAction");
		addMember(computeAction);
		
		in0= new InputTerminal(FleetGlobal.IN+"0");
		terminalNameMap.put(FleetGlobal.IN+"0", in0);
		addMember(in0);
		computeAction.attach(in0);
		
		ackIn0= new OutputTerminal(FleetGlobal.ACK_IN+"0");
		terminalNameMap.put(FleetGlobal.ACK_IN+"0", ackIn0);
		addMember(ackIn0);
		computeAction.attach(ackIn0);

		in1= new InputTerminal(FleetGlobal.IN+"1");
		terminalNameMap.put(FleetGlobal.IN+"1", in1);
		addMember(in1);
		computeAction.attach(in1);
		
		ackIn1= new OutputTerminal(FleetGlobal.ACK_IN+"1");
		terminalNameMap.put(FleetGlobal.ACK_IN+"1", ackIn1);
		addMember(ackIn1);
		computeAction.attach(ackIn1);
		
		tokenOut= new OutputTerminal(FleetGlobal.TOKEN_OUT);
		addMember(tokenOut);
		computeAction.attach(tokenOut);
		
		ackTokenOut= new InputTerminal(FleetGlobal.ACK_TOKEN_OUT);
		addMember(ackTokenOut);
		computeAction.attach(ackTokenOut);

		out= new OutputTerminal(FleetGlobal.OUT);
		addMember(out);
		computeAction.attach(out);
		
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT);
		addMember(ackOut);
		computeAction.attach(ackOut);
		
		tokenIn= new InputTerminal(FleetGlobal.TOKEN_IN);
		addMember(tokenIn);
		computeAction.attach(tokenIn);
		
		ackTokenIn= new OutputTerminal(FleetGlobal.ACK_TOKEN_IN);
		addMember(ackTokenIn);
		computeAction.attach(ackTokenIn);

	}
	
	/** get the terminal name */
	public Terminal getTerminal(String name) {
		// use the map, this is why we made it
		return terminalNameMap.get(name);
	}
	
	
	protected abstract Data compute(Payload a, Payload b);
	
	public String stateInfo() {
		return getName() + ", " + getID() + ": state not defined";
	} // stateInfo
	
	protected class ArithmeticAction extends Action {
		
		public ComputeData computeDataFn;
		public AckFunction ackFn0;
		public AckFunction ackFn1;
		public AckFunction tokenOutFn;
		public AckFunction ackTokenInFn;
		
		public ArithmeticAction(String name) {
			super(name);
			computeDataFn= new ComputeData();
			ackFn0= new AckFunction();
			ackFn1= new AckFunction();
			tokenOutFn= new AckFunction();
			ackTokenInFn= new AckFunction();
			
			// get default delays for tokens and ack tokens
			Integer fDelay= globals.intValue(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_TOKEN_DELAY);
			Delay tokenDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: TokenFifoShipDefaults.TOKEN_FIFO_SHIP_TOKEN_DELAY_DEF;			
			tokenOutFn.setDelay(tokenDelay);	

			fDelay= globals.intValue(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_TOKEN_DELAY);
			Delay ackTokenDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_TOKEN_DELAY_DEF;			
			ackTokenInFn.setDelay(ackTokenDelay);	
			
			outputFunctions.put(FleetGlobal.OUT, computeDataFn);
			outputFunctions.put(FleetGlobal.ACK_IN+"0", ackFn0);			
			outputFunctions.put(FleetGlobal.ACK_IN+"1", ackFn1);			
			outputFunctions.put(FleetGlobal.TOKEN_OUT, tokenOutFn);			
			outputFunctions.put(FleetGlobal.ACK_TOKEN_IN, ackTokenInFn);			
		}
		
		/** Simply pass the data from input terminal 0 */
		class ComputeData extends Function {
			
			public ComputeData() {
				super("PassData");
				Integer fDelay= globals.intValue(ArithmeticShipDefaults.TOKEN_ARITHMETIC_COMPUTE_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: ArithmeticShipDefaults.COMPUTE_DELAY_DEF;
			}
			
			public Object value() {
				// compute data
				Object oa0= in0.getData();
				Object oa1= in1.getData();
				Data result= new Data(null);
				if (oa0 instanceof Payload && oa1 instanceof Payload) {
					Payload p0= (Payload)oa0;
					Payload p1= (Payload)oa1;
					result= compute(p0, p1);
				}
				else {
					if (!(oa0 instanceof Payload)) {
						director.fatalError("Component " + getName() + ", id= "
								+ getID() 
								+ ": non payload data receoived on terminal "
								+ in0.getAlias()
								+ ": " + oa0);
					}
					if (!(oa1 instanceof Payload)) {
						director.fatalError("Component " + getName() + ", id= "
								+ getID() 
								+ ": non payload data receoived on terminal "
								+ in1.getAlias()
								+ ": " + oa1);
					}
				}
				// get data from the input and pass it on
				return result;

					
			} // value
		} // ComputeData		
		
		/** Vacuous function used to acknowldge that the action has fired */ 
		class AckFunction extends Function {
			
			public AckFunction() {
				super("AckFunction");
				Integer fDelay= globals.intValue(ArithmeticShipDefaults.TOKEN_ARITHMETIC_ACK_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: ArithmeticShipDefaults.ACK_DELAY_DEF;
			}
		
			public Data value() {
				return new Data(null);
			} // value
		} // AckFunction
		
	} // class ArithmeticAction
} // class ArithmeticTokenShip
